/*-
 * Copyright (c) 2018 Andrey V. Elsukov <ae@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define TEST_PORT	"44444"

static void
usage(const char *name)
{

	printf("Usage: %s <ipv6-address> [port]\n", name);
	exit(0);
}

static int
ipc_send(int fd, const char *buf)
{

	return (send(fd, buf, strlen(buf) + 1, 0));
}

static int
ipc_recv(int fd, char *buf, size_t len, struct timeval *ptv)
{
	fd_set fds;
	int error;

	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	error = select(fd + 1, &fds, NULL, NULL, ptv);
	if (error < 0)
		err(EX_OSERR, "select");
	if (error == 0)
		return (-1);
	error = recv(fd, buf, len, 0);
	if (error < 0)
		err(EX_OSERR, "ipc_recv");
	return (error);
}

static int
child(int ipc, int sock, const struct addrinfo *a)
{
	char buf[100], ctrl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
	struct timeval tv;
	struct iovec iov;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct in6_pktinfo *pi;
	struct sockaddr_in6 sin;
	int error, ret;

	tv.tv_sec = 5;
	tv.tv_usec = 0;
	ret = 0;

	memcpy(&sin, a->ai_addr, sizeof(sin));
	sin.sin6_addr = in6addr_any;
	sin.sin6_scope_id = 0;
	error = bind(sock, (struct sockaddr *)&sin, sizeof(sin));

	/* Wait READY from parent. */
	if (ipc_recv(ipc, buf, sizeof(buf), &tv) < 0) {
		printf("CHILD: Waiting for READY timed out.\n");
		ret = 1;
	} else {
		printf("CHILD: Received %s from parent.\n", buf);

		msg.msg_flags = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_name = a->ai_addr;
		msg.msg_namelen = a->ai_addrlen;
		msg.msg_control = ctrl;
		msg.msg_controllen = sizeof(ctrl);

		cmsg = CMSG_FIRSTHDR(&msg);
		cmsg->cmsg_level = IPPROTO_IPV6;
		cmsg->cmsg_type = IPV6_PKTINFO;
		cmsg->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		pi = (struct in6_pktinfo *)CMSG_DATA(cmsg);
		memcpy(&sin, a->ai_addr, sizeof(sin));
		pi->ipi6_addr = sin.sin6_addr;
		pi->ipi6_ifindex = sin.sin6_scope_id;

		printf("CHILD: pktinfo [%s]%%%d\n", inet_ntop(AF_INET6,
		    &pi->ipi6_addr, buf, sizeof(buf)), pi->ipi6_ifindex);

		snprintf(buf, sizeof(buf), "TEST DATA");
		iov.iov_base = buf;
		iov.iov_len = strlen(buf) + 1;

		error = sendmsg(sock, &msg, 0);
		if (error < 0)
			err(EX_OSERR, "CHILD: sendmsg");

		error = ipc_send(ipc, "FINISHED");
		if (error < 0)
			err(EX_OSERR, "CHILD: ipc_send");

		getnameinfo(a->ai_addr, a->ai_addrlen,
		    buf, sizeof(buf), NULL, 0, NI_NUMERICHOST);
		printf("CHILD: data sent to [%s]:%d successfully\n", buf,
			ntohs(((struct sockaddr_in6 *)a->ai_addr)->sin6_port));
		printf("CHILD: pktinfo [%s]%%%d\n", inet_ntop(AF_INET6,
		    &pi->ipi6_addr, buf, sizeof(buf)), pi->ipi6_ifindex);
	}
	close(ipc);
	return (ret);
}

static int
parent(int ipc, int sock, const struct addrinfo *a)
{
	char buf[100], ctrl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
	struct sockaddr_in6 sin;
	struct iovec iov;
	struct msghdr msg;
	struct timeval tv;
	struct in6_pktinfo *pi;
	struct cmsghdr *cmsg;
	fd_set fds;
	int error, ret;

	ret = 1;
	error = setsockopt(sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &ret,
	    sizeof(ret));
	if (error < 0)
		err(EX_OSERR, "PARENT: setsockopt");

	tv.tv_sec = 5;
	tv.tv_usec = 0;
	ret = 0;

	memcpy(&sin, a->ai_addr, sizeof(sin));
	sin.sin6_addr = in6addr_any;
	sin.sin6_scope_id = 0;
	error = bind(sock, (struct sockaddr *)&sin, sizeof(sin));
	if (error < 0)
		err(EX_OSERR, "PARENT: bind");

	/* Notify child that we are ready to start testing */
	error = ipc_send(ipc, "READY");
	if (error < 0)
		err(EX_OSERR, "PARENT: ipc_send");

	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	error = select(sock + 1, &fds, NULL, NULL, &tv);
	if (error < 0)
		err(EX_OSERR, "PARENT: select");
	if (error == 0) {
		printf("PARENT: Test has been failed: no data "
		    "received from child\n");
		ret = 1;
	} else {
		iov.iov_base = buf;
		iov.iov_len = sizeof(buf);

		memset(&sin, 0, sizeof(sin));
		msg.msg_name = &sin;
		msg.msg_namelen = sizeof(sin);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = ctrl;
		msg.msg_controllen = sizeof(ctrl);
		msg.msg_flags = 0;

		memset(ctrl, 0, sizeof(ctrl));
		error = recvmsg(sock, &msg, 0);
		if (error < 0)
			err(EX_OSERR, "PARENT: recvmsg");
		printf("PARENT: Received %d bytes from child: %s\n",
		    error, buf);
		error = getnameinfo((const struct sockaddr *)&sin,
		    sizeof(sin), buf, sizeof(buf), NULL, 0, NI_NUMERICHOST);
		if (error == 0) {
			printf("PARENT: The data was received from [%s]:%d\n",
			    buf, ntohs(sin.sin6_port));
			cmsg = CMSG_FIRSTHDR(&msg);
			if (cmsg->cmsg_level == IPPROTO_IPV6 &&
			    cmsg->cmsg_type == IPV6_PKTINFO) {
				pi = (struct in6_pktinfo *)CMSG_DATA(cmsg);
				printf("PARENT: pktinfo [%s]%%%d\n",
				    inet_ntop(AF_INET6, &pi->ipi6_addr, buf,
				    sizeof(buf)), pi->ipi6_ifindex);
			}
			ret = 0;
		} else {
			error = ipc_recv(ipc, buf, sizeof(buf), &tv);
			if (error == 0)
				printf("PARENT: Test has been failed: the "
				    "child reported that it sent data, but "
				    "no data received.\n");
			ret = 2;
		}
	}
	close(ipc);
	return (ret);
}

int
main(int argc, char *argv[])
{
	struct addrinfo hints, *res;
	const char *address, *port;
	pid_t pid;
	int error, sock, pair[2];

	if (argc < 2)
		usage(argv[0]);

	address = argv[1];
	port = argc >= 3 ? argv[2]: TEST_PORT;

	/* Validate the address */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET6;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICHOST;
	error = getaddrinfo(address, port, &hints, &res);
	if (error != 0)
		errx(EX_USAGE, "%s", gai_strerror(error));

	error = socketpair(AF_UNIX, SOCK_DGRAM, 0, pair);
	if (error != 0)
		err(EX_OSERR, "socketpair");

	pid = fork();
	if (pid < 0)
		err(EX_OSERR, "fork");

	/* Create client and server socket for testing */
	sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (sock < 0)
		err(EX_OSERR, "socket");

	if (pid == 0) {
		close(pair[0]);
		error = child(pair[1], sock, res);
	} else {
		close(pair[1]);
		error = parent(pair[0], sock, res);
	}

	freeaddrinfo(res);
	close(sock);
	return (error);
}
